from core.apps import CoreConfig
from django.apps import apps
from core.models import User, InteractiveUser, Officer, UserRole
from django.core.exceptions import PermissionDenied, ValidationError
import logging
from gettext import gettext as _

logger = logging.getLogger(__file__)


def create_or_update_interactive_user(user_id, data, audit_user_id, connected):
    i_fields = {
        "username": "login_name",
        "other_names": "other_names",
        "last_name": "last_name",
        "phone_number": "phone",
        "email": "email",
        "language": "language_id",
        "health_facility_id": "health_facility_id",
    }
    data_subset = {v: data.get(k) for k, v in i_fields.items()}
    data_subset["audit_user_id"] = audit_user_id
    data_subset["role_id"] = data["roles"][0]  # The actual roles are stored in their own table
    data_subset["is_associated"] = connected
    if user_id:
        # TODO we might want to update a user that has been deleted. Use Legacy ID ?
        i_user = InteractiveUser.objects.filter(validity_to__isnull=True, user__id=user_id).first()
    else:
        i_user = InteractiveUser.objects.filter(validity_to__isnull=True, login_name=data_subset["login_name"]).first()
    if i_user:
        i_user.save_history()
        [setattr(i_user, k, v) for k, v in data_subset.items()]
        if "password" in data:
            i_user.set_password(data["password"])
        created = False
    else:
        i_user = InteractiveUser(**data_subset)
        if "password" in data:
            i_user.set_password(data["password"])
        else:
            # No password provided for creation, will have to be set later.
            i_user.stored_password = "locked"
        created = True

    i_user.save()
    create_or_update_user_roles(i_user, data["roles"], audit_user_id)
    if "districts" in data:
        create_or_update_user_districts(
            i_user, data["districts"], data_subset["audit_user_id"]
        )
    return i_user, created


def create_or_update_user_roles(i_user, role_ids, audit_user_id):
    from core import datetime

    now = datetime.datetime.now()
    UserRole.objects.filter(user=i_user, validity_to__isnull=True).update(
        validity_to=now
    )
    for role_id in role_ids:
        UserRole.objects.create(
            user=i_user, role_id=role_id, audit_user_id=audit_user_id
        )


# TODO move to location module ?
def create_or_update_user_districts(i_user, district_ids, audit_user_id):
    # To avoid a static dependency from Core to Location, we'll dynamically load this class
    user_district_class = apps.get_model("location", "UserDistrict")
    from core import datetime

    now = datetime.datetime.now()
    user_district_class.objects.filter(user=i_user, validity_to__isnull=True).update(
        validity_to=now.to_ad_datetime()
    )
    for district_id in district_ids:
        user_district_class.objects.update_or_create(
            user=i_user,
            location_id=district_id,
            defaults={"validity_to": None, "audit_user_id": audit_user_id},
        )


def create_or_update_officer_villages(officer, village_ids, audit_user_id):
    # To avoid a static dependency from Core to Location, we'll dynamically load this class
    officer_village_class = apps.get_model("location", "OfficerVillage")
    from core import datetime

    now = datetime.datetime.now()
    officer_village_class.objects.filter(
        officer=officer, validity_to__isnull=True
    ).update(validity_to=now)
    for village_id in village_ids:
        officer_village_class.objects.update_or_create(
            officer=officer,
            location_id=village_id,
            defaults={"validity_to": None, "audit_user_id": audit_user_id},
        )


def create_or_update_officer(user_id, data, audit_user_id, connected):
    officer_fields = {
        "username": "code",
        "other_names": "other_names",
        "last_name": "last_name",
        "phone_number": "phone",
        "email": "email",
        "birth_date": "dob",
        "address": "address",
        "works_to": "works_to",
        "location_id": "location_id",
        # TODO veo_code, last_name, other_names, dob, phone
        "substitution_officer_id": "substitution_officer_id",
        "phone_communication": "phone_communication",
    }
    data_subset = {v: data.get(k) for k, v in officer_fields.items()}
    data_subset["audit_user_id"] = audit_user_id
    data_subset["has_login"] = connected
    if user_id:
        # TODO we might want to update a user that has been deleted. Use Legacy ID ?
        officer = Officer.objects.filter(
            validity_to__isnull=True, user__id=user_id
        ).first()
    else:
        officer = Officer.objects.filter(
            code=data_subset["code"], validity_to__isnull=True
        ).first()

    if officer:
        officer.save_history()
        [setattr(officer, k, v) for k, v in data_subset.items()]
        created = False
    else:
        officer = Officer(**data_subset)
        created = True

    officer.save()
    if data.get("village_ids"):
        create_or_update_officer_villages(
            officer, data["village_ids"], data_subset["audit_user_id"]
        )
    return officer, created


def create_or_update_claim_admin(user_id, data, audit_user_id, connected):
    ca_fields = {
        "username": "code",
        "other_names": "other_names",
        "last_name": "last_name",
        "phone_number": "phone",
        "email": "email_id",
        "birth_date": "dob",
        "health_facility_id": "health_facility_id",
    }
    data_subset = {v: data.get(k) for k, v in ca_fields.items()}
    data_subset["audit_user_id"] = audit_user_id
    data_subset["has_login"] = connected
    # Since ClaimAdmin is not in the core module, we have to dynamically load it.
    # If the Claim module is not loaded and someone requests a ClaimAdmin, this will raise an Exception
    claim_admin_class = apps.get_model("claim", "ClaimAdmin")
    if user_id:
        # TODO we might want to update a user that has been deleted. Use Legacy ID ?
        claim_admin = claim_admin_class.objects.filter(validity_to__isnull=True, user__id=user_id).first()
    else:
        claim_admin = claim_admin_class.objects.filter(code=data_subset["code"], validity_to__isnull=True).first()

    if claim_admin:
        claim_admin.save_history()
        [setattr(claim_admin, k, v) for k, v in data_subset.items()]
        created = False
    else:
        claim_admin = claim_admin_class(**data_subset)
        created = True

    # TODO update municipalities, regions
    claim_admin.save()
    return claim_admin, created


def create_or_update_core_user(user_uuid, username, i_user=None, t_user=None, officer=None, claim_admin=None):
    if user_uuid:
        # This intentionally fails if the provided uuid doesn't exist as we don't want clients to set it
        user = User.objects.get(id=user_uuid)
        # There is no history to save for User
        if user.username != username:
            logger.warning("Ignored attempt to change the username of %s from %s to %s. This is not supported",
                           user_uuid, user.username, username)
        created = False
    elif username:
        user = User.objects.filter(username=username).first()
        created = False
    else:
        user = None
        created = False

    if not user:
        user = User(username=username)
        created = True

    if i_user:
        user.i_user = i_user
    if t_user:
        user.t_user = t_user
    if officer:
        user.officer = officer
    if claim_admin:
        user.claim_admin = claim_admin
    user.save()
    return user, created


def change_user_password(logged_user, username_to_update=None, old_password=None, new_password=None):
    if username_to_update and username_to_update != logged_user.username:
        if not logged_user.has_perms(CoreConfig.gql_mutation_update_users_perms):
            raise PermissionDenied("unauthorized")
        user_to_update = User.objects.get(username=username_to_update)
    else:
        user_to_update = logged_user
        if not old_password or not user_to_update.check_password(old_password):
            raise ValidationError(_("core.wrong_old_password"))

    user_to_update.set_password(new_password)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='openimis-be-core',
    version='v1.3.1',
    packages=find_packages(),
    include_package_data=True,
    license='GNU AGPL v3',
    description='The openIMIS Backend Core reference module.',
    # long_description=README,
    url='https://openimis.org/',
    author='Xavier Gillmann',
    author_email='xgillmann@bluesquarehub.com',
    install_requires=[
        'django',
        'django-db-signals',
        'djangorestframework',
        'cached-property',
        'nepalicalendar',
        'django-simple-history',
        'django-dirtyfields',
        'websocket-client'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import queue

from enum import Enum

from django import dispatch
from typing import Callable


class ServiceSignalBindType(Enum):
    BEFORE = 0
    AFTER = 1
    BEFORE_AND_AFTER = 2


class RegisteredServiceSignal:

    def __init__(self, providing_args=None):
        """
        Due to how python manage class prototypes, there's a chance that class definition using register_service_signal
        will not be loaded until first import of given class. Therefore it's likely that binding function responsible
        for loading functions called before/after signal will be invoked before signal register. In this case functions
        are queued and connected to actual signal after the registration. Whether or not registration took place
        is defined by providing_args. If argument is None, signal is considered not registered.
        """
        self.__signal_results = {'before': None, 'after': None}
        self.__connection_queue = {type_: queue.Queue() for type_ in ServiceSignalBindType}
        self.__signals_before = []
        self.__signals_after = []

        if not providing_args:
            self.__before_service_signal = None
            self.__after_service_signal = None
        else:
            self.__before_service_signal = dispatch.Signal(providing_args=providing_args)
            self.__after_service_signal = dispatch.Signal(providing_args=providing_args + ['result'])

    @property
    def connected_signals(self):
        return {
            'before': self.__signals_before,
            'after': self.__signals_after
        }

    def send_signal_before(self, sender, **signal_call_args):
        result = self.before_service_signal.send(sender=sender, **signal_call_args)
        self.__signal_results['before'] = result

    def send_signal_after(self, sender, **signal_call_args):
        result = self.after_service_signal.send(sender=sender, **signal_call_args)
        self.__signal_results['after'] = result

    @property
    def signal_results(self):
        return self.__signal_results

    @property
    def before_service_signal(self):
        return self.__before_service_signal

    @property
    def after_service_signal(self):
        return self.__after_service_signal

    def connect_signal(self, func: Callable, bind_type: ServiceSignalBindType = ServiceSignalBindType.BEFORE_AND_AFTER):
        if self.is_signal_registered():
            self._add_connection(func, bind_type)
        else:
            self.__connection_queue[bind_type].put(func)

    def register_signal(self, providing_args):
        if self.is_signal_registered():
            raise ValueError(f"Signal {self} already registered.")

        self.__before_service_signal = dispatch.Signal(providing_args=providing_args)
        self.__after_service_signal = dispatch.Signal(providing_args=providing_args + ['result'])
        self._connect_queued()

    def is_signal_registered(self):
        return self.__before_service_signal is not None and self.__after_service_signal is not None

    def _add_connection(self, func, bind_type):
        if bind_type == ServiceSignalBindType.BEFORE:
            self.__before_service_signal.connect(func)
            self.__signals_before.append(func)
        elif bind_type == ServiceSignalBindType.AFTER:
            self.__after_service_signal.connect(func)
            self.__signals_after.append(func)
        elif bind_type == ServiceSignalBindType.BEFORE_AND_AFTER:
            self.__before_service_signal.connect(func)
            self.__after_service_signal.connect(func)
            self.__signals_after.append(func)
            self.__signals_before.append(func)
        else:
            raise AttributeError(f"Invalid bind_type {bind_type}, should be one of {ServiceSignalBindType}.")

    def _connect_queued(self):
        for binding_type, awaiting_queue in self.__connection_queue.items():
            while not awaiting_queue.empty():
                self.connect_signal(awaiting_queue.get(), binding_type)



# from graphql_auth.backends import GraphQLAuthBackend
# from django.utils.deprecation import MiddlewareMixin
import jwt
from graphql_jwt.settings import jwt_settings
from django.apps import apps

import logging

logger = logging.getLogger(__file__)


def jwt_encode_user_key(payload, context=None):
    token = jwt.encode(
        payload,
        get_jwt_key(encode=True, context=context),
        jwt_settings.JWT_ALGORITHM,
    )
    # JWT module after 1.7 does the encoding, introducing some conflicts in graphql-jwt, let's support both
    if isinstance(token, bytes):
        token = token.decode("utf-8")
    return token


def jwt_decode_user_key(token, context=None):
    # First decode the token without validating it, so we can extract the username
    not_validated = jwt.decode(
        token,
        get_jwt_key(encode=False, context=context),
        options={
            'verify_exp': jwt_settings.JWT_VERIFY_EXPIRATION,
            'verify_aud': jwt_settings.JWT_AUDIENCE is not None,
            'verify_signature': False,
        },
        leeway=jwt_settings.JWT_LEEWAY,
        audience=jwt_settings.JWT_AUDIENCE,
        issuer=jwt_settings.JWT_ISSUER,
        algorithms=[jwt_settings.JWT_ALGORITHM],
    )
    if not_validated and not_validated.get("username"):
        user_class = apps.get_model("core", "User")
        db_user = user_class.objects\
            .filter(username=not_validated.get("username"))\
            .only("i_user__private_key")\
            .first()
        if db_user and db_user.private_key:
            key = db_user.private_key
        else:
            key = get_jwt_key(encode=False)
    else:
        key = get_jwt_key(encode=False)
    return jwt.decode(
        token,
        key,
        options={
            'verify_exp': jwt_settings.JWT_VERIFY_EXPIRATION,
            'verify_aud': jwt_settings.JWT_AUDIENCE is not None,
            'verify_signature': jwt_settings.JWT_VERIFY,
        },
        leeway=jwt_settings.JWT_LEEWAY,
        audience=jwt_settings.JWT_AUDIENCE,
        issuer=jwt_settings.JWT_ISSUER,
        algorithms=[jwt_settings.JWT_ALGORITHM],
    )


def get_jwt_key(encode=True, context=None):
    user_key = extract_private_key_from_context(context)
    if user_key:
        return user_key
    if encode:
        if hasattr(jwt_settings, "JWT_PRIVATE_KEY"):
            return jwt_settings.JWT_PRIVATE_KEY
    else:
        if hasattr(jwt_settings, "JWT_PUBLIC_KEY"):
            return jwt_settings.JWT_PUBLIC_KEY
    return jwt_settings.JWT_SECRET_KEY


def extract_private_key_from_context(context):
    if context and context.user and hasattr(context.user, "private_key"):
        return context.user.private_key
    return None



# Generated by Django 2.1.9 on 2019-07-26 07:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='moduleconfiguration',
            name='is_disabled_until',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='moduleconfiguration',
            name='is_exposed',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='moduleconfiguration',
            name='layer',
            field=models.CharField(choices=[('fe', 'frontend'), ('be', 'backend')], default='be', max_length=2),
        ),
    ]



from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .models import User
from .serializers import UserSerializer


class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    # If we don't specify the IsAuthenticated, the framework will look for the core.user_view permission and prevent
    # any access from non-admin users
    permission_classes = [IsAuthenticated]

    @action(detail=False)
    def current_user(self, request):
        serializer = self.get_serializer(request.user, many=False)
        return Response(serializer.data)
